---------------------------------------------------------------------------------
--	DIGITALER SINUS-GENERATOR
--
-- Modulname:			anzeigen_timer.vhd
-- Projektname:		funktionsgenerator.ise
-- Ersteller:			Thomas Brunnengrber
--	Firma:				Privat
-- Erstellt:			26.03.2006
-- Zielsystem:			Spartan 3
-- Tool versions:		XILINX ISE 8.1
-- Revision:			1.0
-- Abhngigkeiten:	taktteiler2c.vhd
-- 
---------------------------------------------------------------------------------
-- Stellt das Timing fr ein LC-Display auf Basis des HD44780 oder ST7066 sicher.
---------------------------------------------------------------------------------

library ieee;
	use ieee.std_logic_1164.all;
	use ieee.numeric_std.all;

entity Anzeigen_Timer is
	port( Takt			: in std_logic;					-- min. 640ns Periode
			reset			: in std_logic;					-- Asynchroner Reset, active high
			start			: in std_logic;					-- Starte Schreib-Sequenz
			busy			: out std_logic := '0';			-- '1' solange LCD beschftigt
			LCD_Enable	: out std_logic := '0'			-- LCD Enable
		 );
end Anzeigen_Timer;

--	Schreib-Sequenz:
--	1. Daten, RS und R/W_ setzen
--	2. T_1 = 640ns warten (1 Takt)
--	3. EN setzen
--	4. T_2 = 640ns warten
--	5. EN zurcknehmen
--	6. T_3 = 128*640ns warten, dann busy zurcknehmen
--	=> nur ein Takt mit <2MHz ntig

architecture arch1 of Anzeigen_Timer is
-- 9 slices, 11 FF, 16 LUTs
	component taktteiler2c is
		generic ( n : integer range 1 to 30 := 2 );	-- Takt-Teilung durch 2^n 
		port ( loeschen : in std_logic;
				 Takt_ein : in std_logic;
				 Takt_aus : out std_logic );
	end component taktteiler2c;
	
	type Zustand_Typ is ( Leerlauf, T_1, T_2, T_3 );
	signal Zustand : Zustand_Typ := Leerlauf;
	signal Teiler_reset : std_logic;
	signal Takt_128 : std_logic;

begin

	Warteschleife : taktteiler2c
		generic map ( n => 8 )	-- Takt-Teilung durch 2^8=256
										-- => nach 128 Takten geht Takt_aus(8) von 0 auf 1  
	   port map ( loeschen => Teiler_reset, Takt_ein => Takt, Takt_aus => Takt_128 );

	fsm1: process ( Takt, reset, start )
		variable T1_starten : std_logic := '0';	-- bei nchstem Takt nach T1 gehen
	begin
		if reset = '1' then
			Zustand <= Leerlauf;
		elsif rising_edge( Takt ) then
			case Zustand is
				when Leerlauf => 
					if T1_starten = '1' then
						Zustand <= T_1;
					end if;
				when T_1 =>
					Zustand <= T_2;
				when T_2 =>
					Zustand <= T_3;
				when T_3 =>
					if Takt_128 = '1' then
						Zustand <= Leerlauf;
						busy <= '0';
						T1_starten := '0';
					end if;
			end case;
		end if;
		if reset = '1' then
			busy <= '0';
			T1_starten := '0';
		elsif start = '1' then
			busy <= '1';
			T1_starten := '1';
		end if;
	end process;
			
	Teiler_reset <= '1' when Zustand = T_2 else '0';
	LCD_Enable   <= '1' when Zustand = T_2 else '0';

end arch1;

--architecture arch2 of Anzeigen_Timer is
-- 9 slices, 12 FF, 14 LUTs
--
--	component taktteiler2c is
--		generic ( n : integer range 1 to 30 := 2 );	-- Takt-Teilung durch 2^n 
--		port ( loeschen : in std_logic;
--				 Takt_ein : in std_logic;
--				 Takt_aus : out std_logic );
--	end component taktteiler2c;
--	
--	signal Zustand : std_logic_vector( 0 to 3 ) := "0001";
--	signal Teiler_reset : std_logic;
--	signal Takt_128 : std_logic;
--
--begin
--
--	Warteschleife : taktteiler2c
--		generic map ( n => 8 )	-- Takt-Teilung durch 2^8=256
--										-- => nach 128 Takten geht Takt_aus(8) von 0 auf 1  
--	   port map ( loeschen => Teiler_reset, Takt_ein => Takt, Takt_aus => Takt_128 );
--
--	fsm1: process ( Takt, reset )
--	begin
--		if reset = '1' then
--			Zustand <= "0001";
--		elsif rising_edge( Takt ) then
--			if Zustand(3) = '1' and Takt_128 = '1' then
--				Zustand(0) <= '1';
--			else
--				Zustand(0) <= '0';			
--			end if;
--			if Zustand(0) = '1' and start = '1' then
--				Zustand(1) <= '1';
--			else
--				Zustand(1) <= '0';			
--			end if;
--			if Zustand(1) = '1' then
--				Zustand(2) <= '1';
--			else
--				Zustand(2) <= '0';			
--			end if;
--			if Zustand(2) = '1' then
--				Zustand(3) <= '1';
--			else
--				Zustand(3) <= '0';			
--			end if;
--		end if;
--	end process;
--	
--	busy <= (not Zustand(0));
--	Teiler_reset <= Zustand(2);
--	LCD_Enable <= Zustand(2);
--
--end arch2;