--------------------------------------------------------------------------------
--	DIGITALER SINUS-GENERATOR
--
-- Modulname:			sinusgenerator.vhd
-- Projektname:		funktionsgenerator.ise
-- Ersteller:			Thomas Brunnengrber
--	Firma:				Privat
-- Erstellt:			05.03.2006
-- Zielsystem:			Spartan 3
-- Tool versions:		XILINX ISE 8.1
-- Revision:			1.0
-- Abhngigkeiten:	Keine
-- 
--------------------------------------------------------------------------------
-- Beschreibung:
--
-- Dieses Modul ist ein generisches Basismodul zur Erstellung von digitalen
-- Funktionsgeneratoren (hier speziell Sinus).
-- 
-- Das erste Viertel einer Sinuswelle wird als Tabelle in einem RAM-Block des
-- Spartan3 abgelegt. Bei einer Breite eines Tabelleneintrags w_tabeintrag von
-- 16 bit ergibt sich damit eine Tabellenlnge l_tab von 1024 Eintrgen pro
--	BRAM.
-- Die Phasendifferenz zwischen zwei Eintrgen betrgt somit 360/4096.
-- Die tatschliche Auflsung betrgt damit etwa 11,3 bit bei einer Sinus-
-- Funktion. Wenn dies nicht ausreicht, mssen die Funktionswerte auf mehrere 
-- BRAMs verteilt werden.
--
-- In den Bereichen 90-180 und 270-360 werden die Adressen invertiert.
-- Im Bereich 180-360 mssen die Funktionswerte "negiert" werden. Wie diese
-- "Negation" aussieht, ist abhngig vom tatschlich verwendeten DAC. Fordert der
-- DAC Zweier-Komplemente, so kann statt dessen eine schnellere und logiksparende
-- Invertierung aller Bits angewandt werden, da die letzten 4 Bit der
-- Tabelleneintrge aufgrund der durch die Tabellenlnge effektiv auf 11,3 bit
-- begrenzten Auflsung letztlich nicht zur Verfgung stehen und daher der durch
-- die simple Invertierung begangene Fehler nicht von Bedeutung ist.
--
-- Bestimmung von Takt und Lnge des Phasenzhlers:
--
--	Taktfrequenz: f_takt
--	Phasenzhler: phase
--	Bitbreite des Phasenzhlers : w_phase bit
--	Lnge des Phasenzhlers: l_phase = 2^w_phase
--	Bei jedem Takt zum Phasenzhler addierter Wert: d_phase
--	=> Kleinste einzustellende Frequenz: f_min = f_takt / l_phase
--	=> Generierte Frequenz: f_gen = f_min * d_phase
--	=> oder d_phase = f_gen / f_min
--	wobei f_takt und l_phase und damit f_min beim Design festzulegende Konstanten sind.
--	Gibt man f_min und f_takt vor, so ergibt sich die Lnge des Phasenzhlers zu
--	=> l_phase = f_min / f_takt
--	Der Maximalwert d_phase_max von d_phase ergibt sich aus der maximal zu generierenden 
--	Frequenz f_gen_max zu
--	=> d_phase_max = f_gen_max / f_min
--
--	Ganz einfaches Beispiel:
--	w_phase = 8 bit
--	=> l_phase = 2^8 = 256 
--	=> phase = 0..255
--	f_takt = 128 Hz
--	=> f_min = 128 Hz / 256 = 0.5 Hz
--	f_gen = 1 Hz soll eingestellt werden
--	=> d_phase = 1 Hz / 0.5 Hz = 2
--
--	Konkretes Beispiel:
--	Forderungen: 
--		f_min = 0.01 Hz
--		f_max = 10 MHz
--		etwa 2-fach Oversampling bei 10 MHz, damit analoge Filterung einfacher wird
-- 	=> f_takt ca. 40 MHz
-- 
--	=> l_phase ca. 4e7 / 0.01 = 4e9 => w_phase = 32 bit => l_phase = 2^32 = 4294967296
--	=> f_takt = l_phase * f_min = 42949672.96 Hz = 4194304 Hz * 32/10 * 32/10 
--	=> Quarz mit 2^22 Hz = 4194304 Hz und zwei nachgeschaltete DCM mit jeweils 
--	CLKFX_MULTIPLY = 32 und CLKFX_DIVIDE = 10
--	Achtung: Wird der DCM mit solch niedriger Eingangs-Frequenz betrieben, drfen die
-- PLL-Ausgnge nicht genutzt werden!
--	Alternativ: 50 MHz Quarz * 17 * 24 / 19 / 25 = 42947368 Hz
--	=> d_phase_max = 1e7 / 0.01 = 1e9 => w_d_phase = 30 bit
--	=> f_gen_max = 2^30 * 0.01 = 10737418.23 Hz (einstellbar)
--
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library UNISIM;
use UNISIM.VComponents.all;

entity sinus_generator is
	generic (
		w_phase : integer := 12;
		w_d_phase : integer := 10;
		w_fktwert : integer := 6 );
	port (
		takt : in std_logic;
		d_phase : in std_logic_vector ( w_d_phase-1 downto 0 );
		neustart : in std_logic;	-- Trigger bei	steigender Flanke
		start_phase : in std_logic_vector ( w_phase-1 downto 0 ); 
				-- bernahme synchron mit takt, wenn neustart = '1'
		fkt_wert : out std_logic_vector ( w_fktwert-1 downto 0 ) -- Tabelleneintrag
);
end entity sinus_generator;

architecture arch1 of sinus_generator is

	constant w_adr : integer := 10; 			-- Breite der Adresse ins BRAM
	constant w_tabeintrag : integer := 16;	-- Breite jedes Tabellen-Eintrags

	signal adr : std_logic_vector ( w_adr-1 downto 0 );	
	signal halbwelle : std_logic;
	signal tab_wert : std_logic_vector ( w_tabeintrag-1 downto 0 ); 

begin
   -- RAMB16_S18: Virtex-II/II-Pro, Spartan-3/3E 1k x 16 + 2 Parity bits Single-Port RAM
   -- Xilinx  HDL Language Template version 7.1i
   RAMB16_S18_inst : RAMB16_S18
   generic map (
      INIT => X"08000", --  Value of output RAM registers at startup
      SRVAL => X"08000", --  Ouput value upon SSR assertion
      WRITE_MODE => "WRITE_FIRST", --  WRITE_FIRST, READ_FIRST or NO_CHANGE

		INIT_00 => X"030B02D802A602740242020F01DD01AB01780146011400E200AF007D004B0019",
		INIT_01 => X"062E05FC05CA059805650533050104CF049C046A0438040603D403A1036F033D",
		INIT_02 => X"0951091F08ED08BB08880856082407F207C0078E075B072906F706C506930660",
		INIT_03 => X"0C720C400C0E0BDC0BAA0B780B460B140AE20AB00A7E0A4C0A1909E709B50983",
		INIT_04 => X"0F920F600F2E0EFC0ECA0E980E660E340E020DD00D9E0D6C0D3A0D080CD60CA4",
		INIT_05 => X"12AF127D124B121911E811B611841152112110EF10BD108B105910270FF50FC4",
		INIT_06 => X"15C9159715661534150314D1149F146E143C140B13D913A713761344131212E0",
		INIT_07 => X"18E018AE187D184C181A17E917B717861755172316F216C0168F165D162C15FA",
		INIT_08 => X"1BF21BC11B901B5F1B2E1AFD1ACC1A9B1A6A1A381A0719D619A5197319421911",
		INIT_09 => X"1F011ED01EA01E6F1E3E1E0D1DDC1DAB1D7A1D491D181CE81CB71C861C551C24",
		INIT_0A => X"220B21DA21AA21792149211820E820B72087205620251FF51FC41F931F631F32",
		INIT_0B => X"250F24DF24AF247F244F241F23EE23BE238E235E232D22FD22CD229C226C223B",
		INIT_0C => X"280E27DF27AF277F274F271F26EF26C0269026602630260025D025A025702540",
		INIT_0D => X"2B072AD82AA82A792A492A1A29EB29BB298B295C292C28FD28CD289D286E283E",
		INIT_0E => X"2DF92DCA2D9B2D6C2D3D2D0E2CDF2CB02C812C522C232BF32BC42B952B662B36",
		INIT_0F => X"30E430B630873059302A2FFB2FCD2F9E2F6F2F412F122EE32EB52E862E572E28",
		INIT_10 => X"33C73399336B333D330F32E132B332853257322831FA31CC319E316F31413112",
		INIT_11 => X"36A336753648361A35ED35BF359235643536350834DB34AD347F3451342333F5",
		INIT_12 => X"39763949391C38EF38C238953868383B380D37E037B337863758372B36FE36D0",
		INIT_13 => X"3C403C143BE73BBB3B8E3B623B353B093ADC3AAF3A833A563A2939FD39D039A3",
		INIT_14 => X"3F013ED53EA93E7D3E523E263DFA3DCE3DA23D763D493D1D3CF13CC53C993C6C",
		INIT_15 => X"41B8418D41624136410B40E040B54089405E403240073FDB3FB03F843F583F2D",
		INIT_16 => X"4465443B441043E543BB43904365433B431042E542BA428F42644239420E41E3",
		INIT_17 => X"470846DE46B4468A46604636460C45E245B8458D45634539450F44E444BA4490",
		INIT_18 => X"499F4976494D492348FA48D148A8487E4855482B480247D847AE4785475B4731",
		INIT_19 => X"4C2B4C034BDA4BB24B894B614B384B0F4AE74ABE4A954A6C4A434A1A49F149C8",
		INIT_1A => X"4EAC4E844E5C4E344E0D4DE54DBD4D954D6D4D454D1D4CF54CCC4CA44C7C4C54",
		INIT_1B => X"512050F950D250AB5084505D5036500F4FE74FC04F994F714F4A4F224EFB4ED3",
		INIT_1C => X"53885362533B531552EF52C952A2527C5255522F520851E251BB5194516E5147",
		INIT_1D => X"55E355BD55985572554D5528550254DC54B75491546B5445542053FA53D453AE",
		INIT_1E => X"5830580C57E757C3579E577957555730570B56E656C1569C56775652562D5608",
		INIT_1F => X"5A705A4D5A295A0559E159BE599A59765952592E590A58E658C1589D58795855",
		INIT_20 => X"5CA25C805C5D5C3A5C175BF45BD15BAE5B8B5B685B455B215AFE5ADB5AB75A94",
		INIT_21 => X"5EC65EA45E825E605E3F5E1C5DFA5DD85DB65D945D715D4F5D2D5D0A5CE85CC5",
		INIT_22 => X"60DB60BA609960786057603660155FF45FD35FB15F905F6E5F4D5F2B5F0A5EE8",
		INIT_23 => X"62E262C262A26282626162416221620161E061C0619F617F615E613E611D60FC",
		INIT_24 => X"64D964BA649B647B645C643D641E63FE63DF63C063A063806361634163216301",
		INIT_25 => X"66C066A26684666666486629660B65ED65CE65B06591657365546535651664F7",
		INIT_26 => X"6898687B685E68406823680667E967CB67AE6790677367556737671A66FC66DE",
		INIT_27 => X"6A5F6A436A276A0B69EF69D369B6699A697D696169446928690B68EE68D168B5",
		INIT_28 => X"6C166BFB6BE06BC56BAA6B8F6B736B586B3D6B216B066AEA6ACE6AB36A976A7B",
		INIT_29 => X"6DBD6DA36D896D6F6D556D3A6D206D066CEC6CD16CB76C9C6C816C676C4C6C31",
		INIT_2A => X"6F526F396F206F076EEE6ED56EBC6EA36E896E706E576E3D6E246E0A6DF06DD6",
		INIT_2B => X"70D670BF70A7708F7077705F7047702F70166FFE6FE66FCD6FB56F9C6F846F6B",
		INIT_2C => X"72497233721C720571EE71D771C071A97192717B7164714C7135711D710670EE",
		INIT_2D => X"73AB7395737F736A7354733E7328731272FC72E672D072BA72A3728D72767260",
		INIT_2E => X"74FA74E674D174BD74A87493747E746A74557440742B7415740073EB73D673C0",
		INIT_2F => X"76387624761175FD75EA75D675C375AF759B75877573755F754B75377523750F",
		INIT_30 => X"77637751773E772C771A770776F576E276CF76BD76AA769776847671765E764B",
		INIT_31 => X"787C786B78597848783778267814780377F177E077CE77BC77AB779977877775",
		INIT_32 => X"79827972796279527942793279227911790178F178E078CF78BF78AE789D788C",
		INIT_33 => X"7A757A677A587A497A3A7A2B7A1C7A0D79FE79EF79DF79D079C079B179A17992",
		INIT_34 => X"7B567B487B3B7B2D7B1F7B127B047AF67AE87ADA7ACC7ABD7AAF7AA17A927A84",
		INIT_35 => X"7C237C177C0B7BFE7BF27BE57BD97BCC7BBF7BB27BA57B987B8B7B7E7B717B63",
		INIT_36 => X"7CDE7CD37CC87CBC7CB17CA67C9A7C8F7C837C777C6C7C607C547C487C3C7C30",
		INIT_37 => X"7D857D7B7D717D677D5D7D537D497D3E7D347D297D1F7D147D097CFF7CF47CE9",
		INIT_38 => X"7E197E107E077DFF7DF67DED7DE47DDA7DD17DC87DBF7DB57DAC7DA27D987D8F",
		INIT_39 => X"7E997E927E8A7E837E7B7E737E6B7E637E5B7E537E4B7E437E3B7E327E2A7E21",
		INIT_3A => X"7F067F007EF97EF37EED7EE67EDF7ED97ED27ECB7EC47EBD7EB67EAF7EA87EA1",
		INIT_3B => X"7F5F7F5A7F557F507F4B7F457F407F3B7F357F2F7F2A7F247F1E7F187F127F0C",
		INIT_3C => X"7FA57FA17F9D7F997F957F917F8D7F897F857F807F7C7F777F727F6E7F697F64",
		INIT_3D => X"7FD77FD47FD27FCF7FCC7FC97FC67FC37FC07FBD7FBA7FB77FB37FB07FAC7FA9",
		INIT_3E => X"7FF57FF47FF27FF17FEF7FEE7FEC7FEA7FE87FE67FE47FE27FE07FDE7FDC7FD9",
		INIT_3F => X"7FFF7FFF7FFF7FFF7FFF7FFE7FFE7FFD7FFD7FFC7FFB7FFA7FF97FF87FF77FF6",

      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000")
   port map (
      DO => tab_wert,	-- 16-bit Data Output
      DOP => open,		-- 2-bit parity Output
      ADDR => adr,		-- 10-bit Address Input
      CLK => takt,		-- Clock
      DI => X"FFFF",		-- 16-bit Data Input
      DIP => "11",		-- 2-bit parity Input
      EN => '1',			-- RAM Enable Input
      SSR => '0',			-- Synchronous Set/Reset Input
      WE => '0'			-- Write Enable Input
   );

	set_phase : process ( takt )
		variable phase : unsigned ( w_phase-1 downto 0 );
			-- Lnge: nur Viertelwelle in RAM gespeichert => index = 0..4*iadr-1 
			-- => Index-Breite w_itab = w_adr+2
		variable viertelwelle : std_logic;	-- 2. und 4. Viertelwelle => Adresse invertieren

	begin
		if rising_edge( takt ) then
			halbwelle <= std_logic( phase( w_phase-1 ) );
			if neustart = '0' then
				phase := phase + unsigned( d_phase );
			else
				phase := unsigned( start_phase );
			end if;
			viertelwelle := std_logic( phase( w_phase-2 ) );
		end if;

		if viertelwelle = '0' then						-- 1. oder 3. Viertelwelle
			adr <= std_logic_vector( phase( w_phase-3 downto w_phase-w_adr-2 ) );
		else 
			adr <= not( std_logic_vector( phase( w_phase-3 downto w_phase-w_adr-2 ) ) );
		end if;

	end process;

	fkt_wert <= tab_wert( w_tabeintrag-1 downto w_tabeintrag-w_fktwert ) when halbwelle = '0'		-- 1. Halbwelle
			else not( tab_wert( w_tabeintrag-1 downto w_tabeintrag-w_fktwert ) );						-- 2. Halbwelle

end arch1;
