--------------------------------------------------------------------------------
--	DIGITALER SINUS-GENERATOR
--
-- Modulname:			tda1543_interface.vhd
-- Projektname:		funktionsgenerator.ise
-- Ersteller:			Thomas Brunnengrber
--	Firma:				Privat
-- Erstellt:			26.03.2006
-- Zielsystem:			Spartan 3
-- Tool versions:		XILINX ISE 8.1
-- Revision:			0.9
-- Abhngigkeiten:	keine
-- 
--------------------------------------------------------------------------------
-- Beschreibung:
--
-- Dieses Modul bernimmt die Ausgnge der Funktionsgeneratoren (Takt takt_gen),
-- multipliziert mit der jeweiligen Amplitude, addiert den jeweiligen Offset und
-- wandelt das Ergebnis in serielle Ausgangsdaten mit Takt takt_dac (I2S-Format).
--
-- Das MSB des Offset-Werts ist das Vorzeichen-Bit, die weiteren Bit der Betrag.
-- Ist das MSB gesetzt, wird ein "vereinfachtes Zweierkompliment" gebildet,
-- d. h. alle Stellen (auer dem MSB) invertiert, aber nicht 1 addiert. Der 
-- hierdurch gemachte Fehler betrgt 1mV. Falls dieser nicht vernachlssigbar
-- sein sollte, wrde ein weiterer Addierer bentigt.
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.std_logic_arith.all;
use ieee.std_logic_signed.all;

entity tda1543_interface is
	generic ( w_daten : integer := 16;
				 w_amplitude : integer := 11 );
	port ( takt_gen : in std_logic;	-- Takt des Funktionsgenerators (ca. 43 MHz)
												-- wird fr die Synchronisation bentigt
			 takt_dac : in std_logic;	-- Takt fr die Ausgabe an den DAC (ca. 9 MHz)
          reset : in std_logic;
			 kanal1, kanal2 : in std_logic_vector( w_daten-1 downto 0 );
			 a1, a2 : in std_logic_vector( w_amplitude-1 downto 0 );	-- unsigned
			 o1, o2 : in std_logic_vector( w_amplitude downto 0 );	
					-- MSB = Vorzeichen, andere Bits Betrag => KEIN Zweier-Komplement
          ws : out std_logic;
	test_da_daten_k1 : out std_logic_vector( w_daten-1 downto 0 );
          data : out std_logic );
end entity tda1543_interface;

architecture a2 of tda1543_interface is
	signal shiftreg : std_logic_vector( w_daten-1 downto 0 );
		-- shiftreg wird synchron zu takt_dac parallel geladen und synchron zu
		-- takt_dac geschoben.
	signal mult_aus : signed( w_daten+w_amplitude downto 0 );
		-- Ausgang des Multiplizierers.
	signal mult_aus_relevant : signed ( w_daten-1 downto 0 );
		-- Relevante Bits des Ausgangs des Multiplizierers.
	signal da_daten : std_logic_vector( w_daten-1 downto 0 );
		-- Daten werden mit Takt des Schieberegisters takt_dac parallel in das
		--	Schieberegister shiftreg kopiert.
		-- Bei takt_gen=43 MHz wird der Multiplizierer sptestens nach etwa
		-- 25ns geladen, aufgrund von takt_dac = 9 MHz bleiben dann noch ber 
		-- 80ns Setup-Zeit zur Multiplikation, Addition und zur bernahme in
		-- das Schieberegister.
	signal lade_mult : std_logic;
		-- Eingangsregister des Multiplizierers mit anderem Kanal laden.
		-- Wird synchron zu takt_dac gesetzt, wenn ein Kanal bis auf ein Bit
		-- ausgegeben wurde und asynchron rckgesetzt, wenn der Multiplizierer
		-- geladen wird.
		-- Daten werden vom Funktionsgenerator mit dessen Takt takt_gen bernommen.
	signal mult_enable : std_logic;
		-- Bei '1' wird der Multiplizierer bei nchstem takt_gen mit anderem
		-- Kanal geladen. Wird synchron zu takt_gen gesetzt und gelscht. 
	signal kanal : std_logic;
		-- Kanal, der gerade seriell ausgegeben wird (0/1).
	signal zaehler : integer range 0 to w_daten-1;
		-- n_bits-1 Takte:
		-- Bei zaehler = 0 wird das Schieberegister geladen, das MSB erscheint
		-- sofort als data.
		-- Bei zaehler = 1..n_bits-1 wird geschoben, das nchste Bit erscheint als data.
		-- Bei zaehler = n_bits-1 wird der Befehl zum Laden des Multiplizierers
		-- lade_mult gegeben und der Kanal (ws) gendert.
	signal k1_s, k2_s : signed ( w_daten-1 downto 0 );
	signal a1_s, a2_s : signed ( w_amplitude downto 0 );
	signal offs : std_logic_vector ( w_amplitude downto 0 );
	signal o_s : signed ( w_amplitude downto 0 );

begin
	steuerung : process( takt_dac, reset, mult_enable )
 	begin
		if reset = '1' then
			kanal <= '0';
			zaehler <= 0;
			lade_mult <= '0';
		elsif rising_edge( takt_dac ) then
			if zaehler = w_daten-1 then
				lade_mult <= '1';
				if kanal = '0' then
					kanal <= '1';
				else
					kanal <= '0';
				end if;
				zaehler <= 0;
			else
				zaehler <= zaehler + 1;
			end if;
		end if;
		if mult_enable = '1' then
			lade_mult <= '0';
		end if;
	end process steuerung;
	ws <= kanal;

	sr : process ( takt_dac, reset ) 
	begin
		if reset = '1' then
			shiftreg <= ( others => '0' );
		elsif rising_edge(takt_dac) then  
			if zaehler = 0 then 
				shiftreg <= da_daten;
				if kanal = '0' then
					test_da_daten_k1 <= da_daten;
				end if;
			else 
				shiftreg <= shiftreg((w_daten-2) downto 0) & '0';
			end if;
		end if;
	end process sr;
	data <= shiftreg( w_daten-1 );

	mult_strg : process ( reset, takt_gen )
	begin
		if reset = '1' then
			mult_enable <= '0';
		elsif rising_edge( takt_gen )	then
			if lade_mult = '1' then
				mult_enable <= '1';
			else
				mult_enable <= '0';
			end if;
		end if;
	end process mult_strg;

	offs <= o1 when kanal = '0' else o2;
	o_s <= signed(offs) when offs(w_amplitude) = '0'
				else signed( offs(w_amplitude) & (not offs(w_amplitude-1 downto 0) ) );
	k1_s <= signed(kanal1);
	k2_s <= signed(kanal2);
	a1_s <= signed('0' & a1);
	a2_s <= signed('0' & a2);

	mult_add : process( takt_gen )
	begin
		if rising_edge( takt_gen ) then
			if mult_enable = '1' then	-- Multiplizierer soll geladen werden
				if kanal = '0' then		-- Kanal 2 wird gerade an DAC geschickt
					mult_aus <= k1_s * a1_s;
				else
					mult_aus <= k2_s * a2_s;
				end if;
			end if;
		end if;
	end process mult_add;

	mult_aus_relevant <= mult_aus( w_daten+w_amplitude-1 downto w_amplitude );
	da_daten(w_daten-1 downto w_daten-w_amplitude-1) <= 
				mult_aus_relevant(w_daten-1 downto w_daten-w_amplitude-1) + o_s;
	da_daten(w_daten-w_amplitude-2 downto 0) <= 
				std_logic_vector( mult_aus_relevant(w_daten-w_amplitude-2 downto 0) );

end a2;
